#' RainFARM Stochastic Precipitation Downscaling
#'
#' @author Jost von Hardenberg, ISAC-CNR, \email{j.vonhardenberg@isac.cnr.it}
#'
#' @description This function implements the RainFARM stochastic precipitation downscaling method.
#' Adapted for climate downscaling and including orographic correction.
#' References: D'Onofrio et al. 2014, J of Hydrometeorology 15, 830-843; Rebora et. al 2006, JHM 7, 724
#' The functions use julia_setup from the JuliaCall library, 
#' which takes a long time only the first time it is called on a machine
#'
#' @import JuliaCall
#' @examples
#' 
#' source("rainfarm.R")
#' source("rfweights.R")
#' 
#' nf <- 8   # Choose a downscaling by factor 8
#' nens <- 3 # Number of ensemble members
#'
#' # create a test array with dimension 8x8 and 20 timesteps
#' # or provide your own read from a netcdf file
#'
#' pr <- rnorm(8*8*20)
#' dim(pr) <- c(8,8,20)
#' lon_mat <- seq(10,13.5,0.5) # could also be a 2d matrix
#' lat_mat <- seq(40,43.5,0.5)
#'
#' # Create a test array of weights
#' ww <- array(1., dim = c(8 * nf, 8 * nf))
#' 
#' # ... or create proper weights using an external fine-scale climatology file
#' #     Specify a weightsfn filename if you wish to save the weights
#' 
#' rd <- rfweights("./worldclim.nc", nf, lon=lon_mat, lat=lat_mat, weightsfn="", fsmooth = TRUE)
#' 
#' # downscale using weights (ww=1. means do not use weights)
#' r <- rainfarm(pr, lon_mat, lat_mat, nf, fsmooth = TRUE, fglob = FALSE, weights = ww, nens = 2, verbose = TRUE)
#' rd <- r[[1]]
#' lon_f <- r[[2]]
#' lat_f <- r[[3]]
#'
#' @export

rfweights <- function(orofile, nf, lon=NULL, lat=NULL, reffile="",
                      weightsfn="", varname="", fsmooth=TRUE) {

  library(JuliaCall)
  library(ncdf4)
  julia_setup(useRCall = FALSE)
  julia_library("RainFARM")
  if ( (reffile == "") & ( (typeof(lon) == "NULL") | (typeof(lat) == "NULL"))) {
     stop("Neither a reference filename nor lon and lat specified!\n")
  }
  if (reffile == "") {
     xvar <- ncdim_def("lon", "degrees_east", lon, longname = "longitude")
     yvar <- ncdim_def("lat", "degrees_north", lat, longname = "latitude")
     tvar <- ncdim_def("time", "days since 01-01-2000", 0., longname = "time",
                       unlim = T)
     my_ncdf <- ncvar_def("grid", "m", list(xvar, yvar, tvar), -999,
                          longname = "grid", prec = "single")
     reffile <- "__reffile.nc"
     ncfile <- nc_create(reffile, my_ncdf, force_v4 = FALSE)
     nc_close(ncfile)
  }

  ww <- julia_call("rfweights", orofile, reffile, as.integer(nf),
                   weightsfn = weightsfn, varname = varname, fsmooth = fsmooth)

  system("rm -f __reffile.nc gridrf_2_*.nc")
  return(ww)
}

rainfarm <- function(pr, lon, lat, nf, fglob = FALSE,
                     fsmooth = TRUE, weights = 1., nens = 1,
                     verbose = FALSE, slope = 0, kmin=1) {
  library(JuliaCall)
  julia_setup(useRCall = FALSE)
  julia_library("RainFARM")

  r <- julia_call("lon_lat_fine", lon, lat, as.integer(nf),
                  need_return = "R")
  lon_f <- r[[1]]
  lat_f <- r[[2]]

  if (slope == 0) {
    r <- julia_call("fft3d", pr, need_return = "R")
    fxp <- r[[1]]
    sx <- julia_call("fitslopex", fxp, kmin = as.integer(kmin),
                     need_return = "R")
  } else {
    sx <- slope
  }

  rd <- list()
  for (j in 1:nens) {
    rd <- c(rd, list(julia_call("rainfarm", pr, sx, as.integer(nf), weights,
                                fglob = fglob, fsmooth = fsmooth,
                                verbose = verbose, need_return = "R")))
  }

  return(c(list(lon_f), list(lat_f), list(rd)))
}
